#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include "string.h"
#include <signal.h>
#include <pthread.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/wait.h>
#include "pct.h"

int shm_id;
int sem_id;
int pct_shm_id;

void signal_semaphore(int sem_id){
        struct sembuf sops = {0, 1, 0};
        semop(sem_id, &sops, 1);
}

void wait_semaphore(int sem_id){
        struct sembuf sops = {0, -1, 0};
        semop(sem_id, &sops, 1);
}

int front = 0;
int rear = -1;

void enqueue(int pid){
        if (rear == maxProcessStructures - 1){
                rear = -1;
        }
        PCB[++rear].pid = pid;
}

int dequeue(){
        int pid = PCB[front++].pid;
        if (front == maxProcessStructures){
                front = 0;
        }
        return pid;
}

volatile sig_atomic_t childReady = 0;

FILE *logfile;

int next_pid = -1;

void scheduleNext(){
        if(front <= rear){
                int next_pid = dequeue();
                printf("scheduling PID %d\n", PCB[next_pid].pid);
        }
}

void sigusr1_handler(int signo){
        printf("recieved sigusr1 signal\n");
         sleep(2);
        childReady = 1;
}

void clean(){ //Function to remove semaphores and shared memory
        if(shmdt(system_clock) == -1){
                perror("shmdtCLOCK");
                exit(EXIT_FAILURE);
        }
        if(shmctl(shm_id, IPC_RMID, NULL) == -1){
                perror("shmctlclock");
                exit(EXIT_FAILURE);
        }

        if(semctl(sem_id, 0, IPC_RMID, 0) == -1){
                perror("semctl");
                exit(EXIT_FAILURE);
        }
        if(shmdt(PCB) == -1){
                perror("shmdtPCB");
                exit(EXIT_FAILURE);
        }
        if(shmctl(pct_shm_id, IPC_RMID, NULL) == -1){
                perror("shmctlpct");
                exit(EXIT_FAILURE);
        }
}

void sigint(){ //Function to clean on Ctrl-C
        clean();
        exit(0);
}

void help(){
        printf("The program description\n");
}

void maxSeconds(int second){
        /*printf("The max seconds before the program terminates is: %d\n", second);*/
}

void logFileName(char *filename){
        /*printf("Logfile name is set as: %s\n", filename);*/
}

int main(int argc, char *argv[]) {

        signal(SIGINT, sigint);
        int option;
        int second = -1;
        char *filename = NULL;
 srand(time(NULL));

        while((option = getopt(argc, argv, "hs:l:")) != -1){
                switch(option){
                        case 'h':
                                help();
                                return 0;
                        case 's':
                                second = atoi(optarg);
                                break;
                        case 'l':
                                filename = optarg;
                                break;
                        default:
                                fprintf(stderr, "Usage: %s [-h] [-s seconds] [-l filename]\n", argv[0]);
                                return 1;
                }
        }

        if(second != 1){
                maxSeconds(second);
        }
        if(logFileName != NULL){
                logFileName(filename);
        }

        signal(SIGINT, sigint);

        char* shmfile = "child.c";
        char* pctfile = "readme";

        key_t clockKey = ftok(shmfile, 65); //Shared memory for the clock
        shm_id = shmget(clockKey, sizeof(struct Clock), IPC_CREAT | 0666);
        if(shm_id < 0){
                perror("shmget");
                exit(1);
        }

        system_clock = (struct Clock *)shmat(shm_id, NULL, 0);
        if(system_clock == (void *) -1){
                perror("shmat");
                return 1;
        }

        union semun{
                int val;
                struct semid_ds *buf;
                unsigned short *array;
                struct seminfo *__buf;

    }argument;

        argument.val = 1;

        char* semfile = "oss.c";
        key_t semkey = ftok(semfile, 65);
        sem_id = semget(semkey, 1, IPC_CREAT | IPC_EXCL | 0666);
        if(sem_id == -1){
                perror("semaphore creation failed");
                exit(EXIT_FAILURE);
        }

        semctl(sem_id, 0, SETVAL, argument);

        key_t pctKey = ftok(pctfile, 65);
        pct_shm_id = shmget(pctKey, sizeof(process_control_table) * maxProcessStructures, IPC_CREAT | 0666);
        if(pct_shm_id < 0){
                perror("shmgetPCT");
        }

        PCB = (process_control_table *)shmat(pct_shm_id, NULL, 0);
        if(PCB == (void *) -1){
        perror("shmatPCT");
                return 1;
        }

        logfile = fopen("logfile.txt", "w");
                if(logfile == NULL){
                        perror("fopen");
                        exit(1);
                }

        //signal(SIGUSR1, sigusr1_handler);

        //if(int i = 0; i < maxProcessStructures; i++){

        for(int i = 0; i < maxProcessStructures; i++){
                PCB->pid = 0;
                PCB->active = 1;
                PCB->burstTime = 30;
                PCB->CPU_time = 3;
                PCB->totalTime = 0;
                PCB->isBlocked = 0;
                PCB->priority = 0;
        }

        //while(1){

                system_clock->seconds++;
                unsigned int random_nano = rand() % 1000;
                system_clock->nanoSeconds += random_nano;

                                pid_t pid = fork();

                        if (pid == 0){
                                //char pid_logical[18];
                                //sprintf(pid_logical, "%d", i);
                                execl("./child", "./child", NULL);
                                perror("exec");
                                exit(EXIT_FAILURE);

                        } else if (pid > 0) {
                                wait_semaphore(sem_id);
                                //PCB->active = 1;
                                //if(next_pid != -1){
                                //PCB->pid = pid;
                                //enqueue(pid);
                                printf("printing pid %d\n", pid);
                                //next_pid = -1;
                                //enqueue(next_pid);

                                //PCB->CPU_time += 1;
                                //PCB->totalTime += 1;
                                //PCB->burstTime = 1;

                                //printf("Parent - Process ID: %d, CPU Time: %d, Total Time in System: %d\n", getpid(), PCB->CPU_time, PCB->totalTime);
                                //signal_semaphore(sem_id);

                                //dequeue();
                                //scheduleNext();
                                //PCB->active = 0;
                                //scheduleNext();
                                printf("parent recieved message from child");
                                sleep(2);

                                //signal_semaphore(sem_id);
                } else {
                        perror("fork");
                        exit(0);
                }
  fclose(logfile);
//        clean();
        return 0;
}





